from typing import List

import torch
import pandas as pd
import ray
from pyarrow.csv import ConvertOptions

from veloce.feature_column import SparseFeatureDef, DenseFeatureDef


class fillna(object):
    def __init__(self, col_selectors: List, value):
        self.col_selectors = col_selectors
        self.value = value

    def __call__(self, data):
        if isinstance(data, pd.DataFrame):
            return self._call_pandas_dataframe(data)
        else:
            raise NotImplementedError

    def _call_pandas_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        values = {col: self.value for col in self.col_selectors}
        return df.fillna(value=values)


class LabelEncoder(object):
    def __init__(self, col_selectors: List):
        self.col_selectors = col_selectors

    def __call__(self, data):
        if isinstance(data, pd.DataFrame):
            return self._call_pandas_dataframe(data)
        else:
            raise NotImplementedError

    def _call_pandas_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.col_selectors] = df[self.col_selectors].astype("category")
        for col in self.col_selectors:
            df[col] = df[col].cat.codes
        return df


class MinMaxScaler(object):
    def __init__(self, col_selector):
        self.col_selector = col_selector

    def __call__(self, data):
        if isinstance(data, pd.DataFrame):
            return self._call_pandas_dataframe(data)
        else:
            raise NotImplementedError

    def _call_pandas_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        df_max = df[self.col_selector].max()
        df_min = df[self.col_selector].min()
        df[self.col_selector] = df[self.col_selector] / (df_max - df_min) + df_min
        return df


class DataLoader(object):
    def __init__(self, path, *, strings_can_be_null=True, rand_seed=2021):
        self.ds = ray.data.read_csv(
            path,
            convert_options=ConvertOptions(strings_can_be_null=strings_can_be_null),
        )
        self.rand_seed = rand_seed
        self.feat_order = ("dense", "sparse")

    def set_sparse_features(
        self,
        feat_names,
        *,
        use_fillna=True,
        use_label_encoder=True,
        embedding_dim=4,
        default_fillna_value="-1",
        batch_format="pandas",
    ):
        self.sparse_feat_names = feat_names
        if use_fillna:
            self.ds = self.ds.map_batches(
                fillna(self.sparse_feat_names, default_fillna_value),
                batch_format=batch_format,
            )
        if use_label_encoder:
            self.ds = self.ds.map_batches(
                LabelEncoder(self.sparse_feat_names), batch_format=batch_format
            )
        self.sparse_defs = (
            self.ds.map_batches(
                SparseFeatureDef(
                    self.sparse_feat_names,
                    embedding_dim=embedding_dim,
                ),
                batch_size=None,
                batch_format=batch_format,
            )
            .to_pandas()
            .to_dict("records")
        )
        return self

    def set_dense_features(
        self,
        feat_names,
        *,
        dim=1,
        use_fillna=True,
        use_minmax_scaler=True,
        default_fillna_value=0,
        batch_format="pandas",
    ):
        self.dense_feat_names = feat_names
        if use_fillna:
            self.ds = self.ds.map_batches(
                fillna(self.dense_feat_names, default_fillna_value),
                batch_format=batch_format,
            )
        if use_minmax_scaler:
            self.ds = self.ds.map_batches(
                MinMaxScaler(self.dense_feat_names), batch_format=batch_format
            )
        self.dense_defs = (
            self.ds.map_batches(
                DenseFeatureDef(self.dense_feat_names, dimension=dim),
                batch_size=None,
                batch_format=batch_format,
            )
            .to_pandas()
            .to_dict("records")
        )
        return self

    def set_label_column(self, label_name):
        self.label_name = label_name
        return self

    def set_features_order(self, order):
        if not order or not all([o in {"dense", "sparse"} for o in order]):
            raise ValueError("Arg order should be given or invalid")
        self.feat_order = order
        return self

    def split(self, valid_split_factor=0.8, return_type="shard_dict"):
        self.valid_split_factor = valid_split_factor
        valid_idx = int(self.ds.count() * valid_split_factor)
        train_dataset, validation_dataset = self.ds.random_shuffle(
            seed=self.rand_seed
        ).split_at_indices([valid_idx])
        self.train_dataset_pipeline = train_dataset.repeat().random_shuffle_each_window(
            seed=self.rand_seed
        )
        self.validation_dataset_pipeline = validation_dataset.repeat()
        if return_type == "tuple":
            return self.train_dataset_pipeline, self.validation_dataset_pipeline
        return {
            "train": self.train_dataset_pipeline,
            "validation": self.validation_dataset_pipeline,
        }

    def _patch_count_attr(self, opts):
        opts.update(
            {
                "count": self.ds.count(),
                "train_set_count": self.ds.count() * self.valid_split_factor,
            }
        )
        return opts

    def gen_torch_dataset_options(self):
        feature_columns = getattr(self, f"{self.feat_order[0]}_feat_names") + getattr(
            self, f"{self.feat_order[1]}_feat_names"
        )
        return self._patch_count_attr(
            {
                "label_column": self.label_name,
                "feature_columns": feature_columns,
                "label_column_dtype": torch.float,
                "feature_column_dtypes": [torch.float] * len(feature_columns),
            }
        )


def gen_dataset_shards(dataset_pipeline, n_shards, locality_hints):
    dataset_shards = [dict() for _ in range(n_shards)]
    for k, v in dataset_pipeline.items():
        shards = v.split(n_shards, equal=True, locality_hints=locality_hints)
        assert len(shards) == n_shards
        for i in range(len(shards)):
            dataset_shards[i][k] = shards[i]
    return dataset_shards



import torch
from torch.optim.optimizer import Optimizer


class FTRL(Optimizer):
    def __init__(self, params, lr=0.1, beta=1.0, l1=0.01, weight_decay=0):
        if not 0.0 < lr:
            raise ValueError("Invalid lr parameter: {}".format(lr))
        if not 0.0 < beta:
            raise ValueError("Invalid beta parameter: {}".format(beta))
        if not 0.0 <= l1:
            raise ValueError("Invalid l1 parameter: {}".format(l1))

        defaults = dict(lr=lr, beta=beta, l1=l1, weight_decay=weight_decay)
        super(FTRL, self).__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure=None):
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]
                if len(state) == 0:
                    state["z"] = torch.zeros_like(p.data)
                    state["n"] = torch.zeros_like(p.data)
                # update z, n
                z, n = state["z"], state["n"]
                sigma = -n.sqrt()
                n.add_(grad ** 2)
                denom = n.sqrt()
                sigma += denom
                sigma /= group["lr"]
                z.add_(grad - sigma * p.data)
                # update weight
                denom += group["beta"]
                denom /= group["lr"]
                denom += group["weight_decay"]
                d = z.sign() * torch.maximum(z.abs() - group["l1"], torch.zeros_like(z))
                p.data = -d / denom
        return loss



import torch
import torch.nn as nn

DEFAULT_GROUP_NAME = "default_group"


class SparseFeat(object):
    key = "sparse"

    def __init__(
        self,
        name,
        vocabulary_size,
        column_idx,
        *,
        embedding_dim=4,
        dtype="int32",
        group_name=DEFAULT_GROUP_NAME,
        feat_type="SparseFeat",
    ):
        if embedding_dim == "auto":
            embedding_dim = 6 * int(pow(vocabulary_size, 0.25))
        self.name = name
        self.vocabulary_size = int(vocabulary_size)
        self.column_idx = int(column_idx)
        self.embedding_dim = int(embedding_dim)
        self.dtype = dtype
        self.group_name = group_name
        self.feat_type = feat_type


class DenseFeat(object):
    key = "dense"

    def __init__(
        self, name, column_idx, *, dimension=1, dtype="float32", feat_type="DenseFeat"
    ):
        self.name = name
        self.column_idx = int(column_idx)
        self.dimension = int(dimension)
        self.dtype = dtype
        self.feat_type = feat_type


def find_feature_values(value):
    try:
        return all(["feat_type" in v for v in value])
    except:
        return False


def rebuild_feature_values(feature_values):
    new_feature_values = list()
    for feat in feature_values:
        feat_variant = globals()[feat["feat_type"]]
        new_feature_values.append(feat_variant(**feat))
    return new_feature_values


def concat_inputs(sparse_embedding_list, dense_value_list):
    if len(sparse_embedding_list) > 0 and len(dense_value_list) > 0:
        sparse_dnn_input = torch.flatten(
            torch.cat(sparse_embedding_list, dim=-1), start_dim=1
        )
        dense_dnn_input = torch.flatten(
            torch.cat(dense_value_list, dim=-1), start_dim=1
        )
        return concatenate([sparse_dnn_input, dense_dnn_input])
    elif len(sparse_embedding_list) > 0:
        return torch.flatten(torch.cat(sparse_embedding_list, dim=-1), start_dim=1)
    elif len(dense_value_list) > 0:
        return torch.flatten(torch.cat(dense_value_list, dim=-1), start_dim=1)
    else:
        raise NotImplementedError


def embedding_dict_gen(
    sparse_feature_defs, init_std=0.0001, linear=False, sparse=False, device="cpu"
):
    embedding_dict = nn.ModuleDict(
        {
            feat.name: nn.Embedding(
                feat.vocabulary_size,
                feat.embedding_dim if not linear else 1,
                sparse=sparse,
            )
            for feat in sparse_feature_defs
        }
    )

    for tensor in embedding_dict.values():
        nn.init.normal_(tensor.weight, mean=0, std=init_std)

    return embedding_dict.to(device)


def compute_inputs_dim(
    sparse_feature_defs=None, dense_feature_defs=None, feature_group=False
):
    input_dim = 0
    if sparse_feature_defs is not None:
        if feature_group:
            sparse_input_dim = len(sparse_feature_defs)
        else:
            sparse_input_dim = sum(feat.embedding_dim for feat in sparse_feature_defs)
        input_dim += sparse_input_dim
    if dense_feature_defs is not None:
        dense_input_dim = sum(map(lambda x: x.dimension, dense_feature_defs))
        input_dim += dense_input_dim
    return input_dim


def collect_inputs_and_embeddings(
    X,
    sparse_feature_defs=None,
    dense_feature_defs=None,
    embedding_layer_def=None,
):
    sparse_feature_defs = sparse_feature_defs or []
    dense_feature_defs = dense_feature_defs or []
    # embeddings part
    if not embedding_layer_def:
        sparse_embeddings = []
    else:
        sparse_embeddings = [
            embedding_layer_def[feat.name](
                X[:, feat.column_idx : feat.column_idx + 1].long()
            )
            for feat in sparse_feature_defs
        ]
    # dense inputs part
    dense_values = [
        X[:, feat.column_idx : feat.column_idx + feat.dimension]
        for feat in dense_feature_defs
    ]
    return dense_values, sparse_embeddings


def concatenate(inputs, axis=-1):
    if len(inputs) == 1:
        return inputs[0]
    else:
        return torch.cat(inputs, dim=axis)



import torch
import torch.nn as nn

from .activation import activation_gen


class DNN(nn.Module):
    def __init__(
        self,
        inputs_dim,
        hidden_units,
        activation="relu",
        l2_reg=0,
        dropout_rate=0,
        use_bn=False,
        init_std=0.0001,
        seed=1024,
        device="cpu",
    ):
        super(DNN, self).__init__()
        if len(hidden_units) == 0:
            raise ValueError("hidden_units is empty!!")

        self.dropout = nn.Dropout(dropout_rate)
        self.seed = seed
        self.l2_reg = l2_reg
        self.use_bn = use_bn
        hidden_units = [inputs_dim] + list(hidden_units)

        # setup layers of dnn network
        self.linear_layers = nn.ModuleList()
        self.activation_layers = nn.ModuleList()
        if self.use_bn:
            self.bn_layers = nn.ModuleList()

        for i in range(len(hidden_units) - 1):
            self.linear_layers.append(nn.Linear(hidden_units[i], hidden_units[i + 1]))
            self.activation_layers.append(activation_gen(activation))
            if self.use_bn:
                self.bn_layers.append(nn.BatchNorm1d(hidden_units[i + 1]))

        for name, tensor in self.linear_layers.named_parameters():
            if "weight" in name:
                nn.init.normal_(tensor, mean=0, std=init_std)

        self.to(device)

    def forward(self, inputs):
        current = inputs
        for i in range(len(self.linear_layers)):
            layer_out = self.linear_layers[i](current)
            if self.use_bn:
                layer_out = self.bn_layers[i](layer_out)
            layer_out = self.activation_layers[i](layer_out)
            layer_out = self.dropout(layer_out)
            current = layer_out
        return current


class OutputLayer(nn.Module):
    def __init__(self, output_fn=None, output_fn_args=None, use_bias=True):
        super(OutputLayer, self).__init__()
        if not output_fn:
            raise ValueError("Arg output_fn must be given")
        if not output_fn_args:
            output_fn_args = dict()

        self.use_bias = use_bias
        self.output_fn = output_fn
        self.output_fn_args = output_fn_args
        if self.use_bias:
            self.bias = nn.Parameter(torch.zeros((1,)))

    def forward(self, X):
        output = X
        if self.use_bias:
            output += self.bias
        output = self.output_fn(output, **self.output_fn_args)
        return output



import inspect

import ray

TIME_DIFF = "epoch_seconds"
EARLY_STOPPED = "is_early_stopped"
CHECKPOINT_BASED = "is_checkpoint_based"


class StyleCoder(object):
    HEADER = "\033[95m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    BOLDCYAN = "\033[1;96m"
    GREEN = "\033[92m"
    BOLDGREEN = "\033[1;92m"
    WARNING = "\033[93m"
    RED = "\033[91m"
    BOLDRED = "\033[1;91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    def __init__(self, use_style):
        self.use_style = use_style

    def __getattr__(self, code):
        if self.use_style:
            try:
                rc = getattr(self, code)
            except:
                raise AttributeError("Arg code is not valid")
        else:
            rc = ""
        return rc


def get_package_name(func):
    return get_module_name(func).split(".")[0]


def get_module_name(func):
    return inspect.getmodule(func).__name__


def get_type(func):
    return type(func).__name__


def get_func_name(func):
    return func.__name__


def inspect_func_args(config, in_func):
    in_func_args = dict()
    requires_arg = list(inspect.signature(in_func.__init__).parameters.keys())
    requires_arg.pop(0)
    for k, v in config.items():
        if k in requires_arg:
            in_func_args[k] = v
    return in_func_args


def merge_results(
    validation_result,
    train_result=None,
    time_diff=None,
    is_early_stopped=False,
    is_checkpoint_based=False,
):
    result = dict(
        is_early_stopped=is_early_stopped, is_checkpoint_based=is_checkpoint_based
    )
    if validation_result is None:
        raise ValueError("Arg validation_result should be given")
    if train_result is None:
        train_result = dict()
    for k, v in train_result.items():
        result[f"train/{k}"] = v
    for k, v in validation_result.items():
        result[f"valid/{k}"] = v
    if time_diff is not None:
        result[TIME_DIFF] = time_diff
    return result


def pprint_results(
    run_results,
    use_style=True,
    show_avg_results=True,
    show_epoch_results=True,
    show_post_analysis=True,
    print_interval=1,
):
    s = StyleCoder(use_style)
    is_es = False
    is_ckpt = False
    for run_idx, worker_results in enumerate(run_results):
        print(f"\n{s.BOLD}Run {run_idx}: {s.ENDC}")
        total = 0
        n_workers = len(worker_results or [])
        avg_metrics = {}
        for worker_idx, results in enumerate(worker_results or []):
            if not len(results):
                continue
            time_diff = results[0].pop(TIME_DIFF)
            is_es = results[0].pop(EARLY_STOPPED) or is_es
            is_ckpt = results[0].pop(CHECKPOINT_BASED) or is_ckpt
            acc_metrics = {k: v for k, v in results[0].items()}
            if len(avg_metrics.keys()) == 0:
                avg_metrics = {k: v for k, v in results[-1].items()}
            else:
                for k, v in results[-1].items():
                    avg_metrics[k] += v
            total = len(results)
            if show_epoch_results:
                print(
                    f"{s.BOLD}========================="
                    f"\nWorker {worker_idx} training results"
                    f"\n========================={s.ENDC}"
                )
            for idx in range(print_interval - 1, total, print_interval):
                val = results[idx]
                if TIME_DIFF in val:
                    time_diff = val.pop(TIME_DIFF)
                if EARLY_STOPPED in val:
                    is_es = val.pop(EARLY_STOPPED) or is_es
                if CHECKPOINT_BASED in val:
                    is_ckpt = val.pop(CHECKPOINT_BASED) or is_ckpt
                metrics_join = []
                for k, v in val.items():
                    if idx == total - 1:
                        acc_metrics[k] -= v
                    metrics_join.append(f"{k}: {'%.5f' % v}")
                metrics_join = "\t".join(metrics_join)
                print(
                    f"[epoch {idx + 1}/{total}: {'%.3f' % time_diff}s]\t{metrics_join}"
                )
            acc_metrics_join = "\t".join(
                [
                    f"{k}: {s.BOLDGREEN}↓ {s.ENDC}{'%.3f' % (v*100)}%"
                    if v > 0
                    else f"{k}: {s.BOLDRED}↑ {s.ENDC}{'%.3f' % -(v*100)}%"
                    for k, v in acc_metrics.items()
                ]
            )
            es_indicator = " ES" if is_es else ""
            ckpt_indicator = " CKPT" if is_ckpt else ""
            if show_post_analysis:
                print(
                    f"{s.BOLD}======================"
                    f"\nWorker {worker_idx} post-analysis"
                    f"\n======================{s.ENDC}"
                    f"\n[epoch 1 {s.BOLD}→{s.ENDC} {total}{s.BOLDCYAN}"
                    f"{ckpt_indicator}{es_indicator}{s.ENDC}]\t{acc_metrics_join}\n"
                )
        if show_avg_results:
            metrics_join = []
            for k, v in avg_metrics.items():
                if k in [EARLY_STOPPED, CHECKPOINT_BASED, TIME_DIFF]:
                    continue
                metrics_join.append(f"{k} avg: {'%.5f' % (v / n_workers)}")
            metrics_join = "\t".join(metrics_join)
            print(
                f"{s.BOLD}========================="
                f"\nRun average final results"
                f"\n========================={s.ENDC}"
                f"\n{metrics_join}\n"
            )


def ticks(iter):
    ops = list(iter)
    for resp in ops:
        ray.get(resp)


def calculate_batches(subset_cnt, size, n):
    batches = subset_cnt / size / n
    if batches.is_integer():
        return batches
    else:
        return int(batches) + 1


def get_batches(config, n_workers):
    return calculate_batches(
        config["dataset_options"]["train_set_count"], config["batch_size"], n_workers
    ), calculate_batches(
        config["dataset_options"]["count"]
        - config["dataset_options"]["train_set_count"],
        config["batch_size"],
        n_workers,
    )



import ray

INIT_FLAG = "ENSCALE_RAY_ENVIRON_INITIALIZED"


def environ_validate(n_cpus=2, n_gpus=0, cluster_endpoint=None, **kwargs):
    ray.init(num_cpus=n_cpus, num_gpus=n_gpus, address=cluster_endpoint, **kwargs)



import torch

from veloce.layer import OutputLayer
from veloce.model.ctr.base import Linear, BaseModel


class LR(BaseModel):
    def __init__(
        self,
        dense_feature_defs=None,
        sparse_feature_defs=None,
        l2_reg_linear=1e-3,
        seed=1024,
        device="cpu",
        output_fn=torch.sigmoid,
        output_fn_args=None,
    ):
        super(LR, self).__init__(seed=seed, device=device)
        self.linear = Linear(
            sparse_feature_defs=sparse_feature_defs,
            dense_feature_defs=dense_feature_defs,
            device=device,
        )
        self.add_regularization_weight(self.linear.parameters(), l2=l2_reg_linear)
        self.output = OutputLayer(output_fn=output_fn, output_fn_args=output_fn_args)

    def forward(self, X):
        logit = self.linear(X)
        return self.output(logit)



# Examples:
# 0.1.0.devN # Developmental release
# 0.1.0aN  # Alpha release
# 0.1.0bN  # Beta release
# 0.1.0rcN  # Release Candidate
# 0.1.0  # Final release
# 0.1.0.postN  # Final post release
__version__ = "0.0.1rc2"



class Arch(object):
    PS = "parameter-server"
    RingAllreduce = "ring-allreduce"


class UpdateStrategy(object):
    Sync = "sync"
    Async = "async"


class HeterogeneousStrategy(object):
    def __init__(
        self,
        dense_arch=Arch.RingAllreduce,
        dense_update_strategy=UpdateStrategy.Sync,
        sparse_arch=Arch.RingAllreduce,
        sparse_update_strategy=UpdateStrategy.Sync,
    ):
        self.dense_arch = dense_arch
        self.dense_update_strategy = dense_update_strategy
        self.sparse_arch = sparse_arch
        self.sparse_update_strategy = sparse_update_strategy

    def use_raytrainer(self):
        if (
            self.dense_arch == Arch.RingAllreduce
            and self.dense_update_strategy == UpdateStrategy.Sync
            and self.sparse_arch == Arch.RingAllreduce
            and self.sparse_update_strategy == UpdateStrategy.Sync
        ):
            return True
        return False

    def use_sync_ps(self):
        if (
            self.dense_arch == Arch.PS
            and self.dense_update_strategy == UpdateStrategy.Sync
            and self.sparse_arch == Arch.PS
            and self.sparse_update_strategy == UpdateStrategy.Sync
        ):
            return True
        return False


class PSStrategy(HeterogeneousStrategy):
    def __init__(self, update_strategy=UpdateStrategy.Sync):
        super(PSStrategy, self).__init__(
            dense_arch=Arch.PS,
            dense_update_strategy=update_strategy,
            sparse_arch=Arch.PS,
            sparse_update_strategy=update_strategy,
        )



import copy

from ray.train import TrainingCallback

ignore_keys = {"_timestamp", "_time_this_iter_s", "_training_iteration"}


class EarlyStoppingCallback(TrainingCallback):
    def __init__(self, trainer):
        self.trainer = trainer
        self.stored_results = []

    def handle_result(self, results):
        for worker_id, origin_worker_result in enumerate(results):
            worker_result = copy.deepcopy(origin_worker_result)
            for key in ignore_keys:
                worker_result.pop(key)
            if len(self.stored_results) != len(results):
                self.stored_results.append([worker_result])
            else:
                self.stored_results[worker_id].append(worker_result)
            if (
                "is_early_stopped" in worker_result
                and worker_result["is_early_stopped"]
            ):
                self.trainer.shutdown()



from veloce.preprocessing import DataLoader

BENCHMARK_DATASET_DEFAULT = "examples/dataset/ctr/criteo_mini.txt"


def load_benchmark_dataset(
    data_path=BENCHMARK_DATASET_DEFAULT,
    *,
    valid_split_factor=0.85,
    rand_seed=2021,
    embedding_dim=4,
    separate_valid_dataset=True,
):
    dataloader = DataLoader(data_path, rand_seed=rand_seed)
    dataloader = (
        dataloader.set_label_column(label_name="label")
        .set_dense_features(feat_names=[f"I{i}" for i in range(1, 14)])
        .set_sparse_features(
            feat_names=[f"C{i}" for i in range(1, 27)], embedding_dim=embedding_dim
        )
        .set_features_order(order=("dense", "sparse"))
    )
    datasets = dataloader.split(valid_split_factor=valid_split_factor)
    torch_dataset_options = dataloader.gen_torch_dataset_options()
    feature_defs = {"dense": dataloader.dense_defs, "sparse": dataloader.sparse_defs}
    if separate_valid_dataset:
        validation_ds = datasets.pop("validation")
        return (datasets, validation_ds), feature_defs, torch_dataset_options
    return datasets, feature_defs, torch_dataset_options



import time
from types import MethodType

import torch
import ray
import numpy as np

from veloce.train_fn import BaseTrainFn
from veloce.inputs import rebuild_feature_values, find_feature_values
from veloce.util import inspect_func_args, ticks, merge_results, get_batches
from veloce.preprocessing import gen_dataset_shards
from veloce.epochvisor import LossMetricAccumulator


# patch funcs for nn modules
def get_weights(self):
    return {k: v.cpu() for k, v in self.state_dict().items()}


def set_weights(self, weights):
    self.load_state_dict(weights)


def get_gradients(self):
    grads = []
    for p in self.parameters():
        grad = None if p.grad is None else p.grad.data.cpu().numpy()
        grads.append(grad)
    return grads


def set_gradients(self, gradients):
    for g, p in zip(gradients, self.parameters()):
        if g is not None:
            p.grad = torch.from_numpy(g)


@ray.remote
class ParameterServer(BaseTrainFn):
    def __init__(self, model, config):
        self.set_config(config)
        for k in config.keys():
            if find_feature_values(config[k]):
                config[k] = rebuild_feature_values(config[k])
        model_args = inspect_func_args(config, model)
        self.setup_model(model_args=model_args, model=model, enable_ddp=False)
        # patch weight & grad funcs
        self.model.get_weights = MethodType(get_weights, self.model)
        self.model.set_gradients = MethodType(set_gradients, self.model)

    def apply_gradients(self, *gradients):
        summed_gradients = [
            np.stack(gradient_zip).sum(axis=0) for gradient_zip in zip(*gradients)
        ]
        self.optimizer.zero_grad()
        self.model.set_gradients(summed_gradients)
        self.optimizer.step()
        return self.model.get_weights()

    def get_weights(self):
        return self.model.get_weights()


@ray.remote
class DataWorker(BaseTrainFn):
    def __init__(self, model, config, n_batches):
        self.n_batches = n_batches
        self.set_config(config)
        for k in config.keys():
            if find_feature_values(config[k]):
                config[k] = rebuild_feature_values(config[k])
        model_args = inspect_func_args(config, model)
        self.setup_model(
            model_args=model_args, model=model, enable_optimizer=False, enable_ddp=False
        )
        # patch weight & grad funcs
        self.model.set_weights = MethodType(set_weights, self.model)
        self.model.get_gradients = MethodType(get_gradients, self.model)

    def consume_dataset_shard(self, dataset_shard):
        train_dataset_shard = dataset_shard["train"]
        self.train_dataset_iter = train_dataset_shard.iter_datasets()
        try:
            validation_dataset_shard = dataset_shard["validation"]
            self.validation_dataset_iter = validation_dataset_shard.iter_datasets()
        except:
            self.validation_dataset_iter = None

    def next_epoch(self, key):
        # set up torch dataset iter
        dataset_iter = iter(
            next(getattr(self, f"{key}_dataset_iter")).to_torch(
                label_column=self.dataset_options["label_column"],
                feature_columns=self.dataset_options["feature_columns"],
                label_column_dtype=self.dataset_options["label_column_dtype"],
                feature_column_dtypes=self.dataset_options["feature_column_dtypes"],
                batch_size=self.batch_size,
            )
        )
        setattr(self, f"{key}_torch_dataset_iter", dataset_iter)
        # set up LossMetricAccumulator
        self.lma = LossMetricAccumulator(
            loss_fn=self.loss_fn, metric_fns=self.metric_fns
        )

    def compute_gradients(self, weights):
        try:
            X, y = next(self.train_torch_dataset_iter)
        except StopIteration:
            self.next_epoch("train")
            X, y = next(self.train_torch_dataset_iter)
        self.model.set_weights(weights)
        self.model.train()
        X = X.to(self.device)
        y = y.to(self.device)
        self.model.zero_grad()
        pred = self.model(X)
        loss = self.loss_fn(pred, y)
        loss.backward()
        self.lma.acc(pred=pred, y=y)
        return self.model.get_gradients()

    def evaluate(self, weights):
        try:
            X, y = next(self.validation_torch_dataset_iter)
        except StopIteration:
            self.next_epoch("validation")
            X, y = next(self.validation_torch_dataset_iter)
        self.model.set_weights(weights)
        self.model.eval()
        with torch.no_grad():
            X = X.to(self.device)
            y = y.to(self.device)
            pred = self.model(X)
            self.lma.acc(pred=pred, y=y)

    def gen_result(self):
        return self.lma.gen_result(num_batches=self.n_batches)


class PSTrainer(object):
    def __init__(self, num_workers, use_gpu):
        self.num_workers = num_workers
        self.use_gpu = use_gpu

    def run(self, model, config, dataset):
        result_dict = {i: [] for i in range(self.num_workers)}
        n_train_batches, n_valid_batches = get_batches(config, self.num_workers)
        iteration = config["epochs"] * n_train_batches
        # TODO: add checkpoint and callback support
        ps = ParameterServer.remote(model, config)
        workers = [
            DataWorker.remote(model, config, n_train_batches)
            for _ in range(self.num_workers)
        ]
        dataset_shards = gen_dataset_shards(
            dataset, n_shards=self.num_workers, locality_hints=workers
        )
        ticks(
            worker.consume_dataset_shard.remote(shard)
            for shard, worker in zip(dataset_shards, workers)
        )
        ticks(worker.next_epoch.remote("train") for worker in workers)
        current_weights = ps.get_weights.remote()
        start_ts = time.time()
        for i in range(iteration):
            # train epoch
            gradients = [
                worker.compute_gradients.remote(current_weights) for worker in workers
            ]
            current_weights = ray.get(ps.apply_gradients.remote(*gradients))
            if i % n_train_batches == 0:
                train_results = ray.get(
                    [worker.gen_result.remote() for worker in workers]
                )
                ticks(worker.next_epoch.remote("validation") for worker in workers)
                # validation epoch
                for _ in range(n_valid_batches):
                    ray.get(
                        [worker.evaluate.remote(current_weights) for worker in workers]
                    )
                validation_results = ray.get(
                    [worker.gen_result.remote() for worker in workers]
                )
                end_ts = time.time()
                for j in range(self.num_workers):
                    result_dict[j].append(
                        merge_results(
                            validation_result=validation_results[j],
                            train_result=train_results[j],
                            time_diff=end_ts - start_ts,
                        )
                    )
                start_ts = time.time()
        return result_dict.values()



from veloce.util import get_module_name


class OptimizerStack(object):
    def __init__(self, *optimizers):
        self.optimizers_stash = optimizers
        self.optimizers = dict()

    def compile(self, model_inst):
        for optim in self.optimizers_stash:
            model_layer = getattr(model_inst, optim["model_key"])
            if "args" not in optim:
                optim["args"] = dict()
            self.optimizers[get_module_name(optim["cls"])] = optim["cls"](
                model_layer.parameters(), **optim["args"]
            )

    def zero_grad(self):
        for optim in self.optimizers.values():
            optim.zero_grad()

    def step(self):
        for optim in self.optimizers.values():
            optim.step()

    def state_dict(self):
        return {k: v.state_dict() for k, v in self.optimizers.items()}

    def load_state_dict(self, state_dict):
        for key, optim in self.optimizers.items():
            optim.load_state_dict(state_dict[key])



import torch

from veloce.layer import OutputLayer, FMNative
from veloce.model.ctr.base import BaseModel
from veloce.inputs import (
    embedding_dict_gen,
    collect_inputs_and_embeddings,
    concat_inputs,
)


class FM(BaseModel):
    def __init__(
        self,
        dense_feature_defs=None,
        sparse_feature_defs=None,
        k_factor=10,
        l2_reg_fm=1e-3,
        l2_reg_embedding=1e-3,
        fm_dropout=0,
        seed=1024,
        device="cpu",
        output_fn=torch.sigmoid,
        output_fn_args=None,
        init_std=1e-4,
    ):
        super(FM, self).__init__(seed=seed, device=device)
        self.dense_defs = dense_feature_defs
        self.sparse_defs = sparse_feature_defs
        self.embedding_layer = embedding_dict_gen(
            sparse_feature_defs=self.sparse_defs,
            init_std=init_std,
            sparse=False,
            device=device,
        )
        self.fm = FMNative(
            feature_def_dims=sum(fc.dimension for fc in self.dense_defs)
            + sum(fc.embedding_dim for fc in self.sparse_defs),
            k_factor=k_factor,
            dropout_rate=fm_dropout,
            init_std=init_std,
        )
        self.add_regularization_weight(
            self.embedding_layer.parameters(), l2=l2_reg_embedding
        )
        self.add_regularization_weight(self.fm.parameters(), l2=l2_reg_fm)
        self.output = OutputLayer(output_fn=output_fn, output_fn_args=output_fn_args)

    def forward(self, X):
        dense_vals, sparse_embs = collect_inputs_and_embeddings(
            X,
            sparse_feature_defs=self.sparse_defs,
            dense_feature_defs=self.dense_defs,
            embedding_layer_def=self.embedding_layer,
        )
        logit = self.fm(concat_inputs(sparse_embs, dense_vals))
        return self.output(logit)



import torch.nn as nn
import torch

from veloce.layer import DNN, OutputLayer, InnerProduct, OuterProduct
from veloce.inputs import (
    concat_inputs,
    compute_inputs_dim,
    embedding_dict_gen,
    collect_inputs_and_embeddings,
    concatenate,
)
from .base import BaseModel


class PNN(BaseModel):
    def __init__(
        self,
        # feature_defs
        dense_feature_defs=None,
        sparse_feature_defs=None,
        # specific config
        use_inner=True,
        use_outer=False,
        outer_kernel_type="mat",
        embedding_size=4,
        # dnn related
        dnn_hidden_units=(256, 128),
        dnn_use_bn=False,
        dnn_activation="relu",
        l2_reg_embedding=1e-3,
        l2_reg_dnn=1e-3,
        dnn_dropout=0,
        # base config
        seed=1024,
        output_fn=torch.sigmoid,
        output_fn_args=None,
        device="cpu",
        init_std=1e-4,
    ):
        super(PNN, self).__init__(seed=seed, device=device)
        if outer_kernel_type not in ["mat", "vec", "num"]:
            raise ValueError("Arg kernel_type must be mat, vec or num")

        self.dense_defs = dense_feature_defs
        self.sparse_defs = sparse_feature_defs
        self.use_dnn = len(dnn_hidden_units) > 0
        self.use_inner = use_inner
        self.use_outer = use_outer
        self.outer_kernel_type = outer_kernel_type
        self.embedding_size = embedding_size

        product_out_dim = 0
        inputs_dim = compute_inputs_dim(self.sparse_defs, feature_group=True)
        num_pairs = int(inputs_dim * (inputs_dim - 1) / 2)

        # product layer setup
        if self.use_inner:
            product_out_dim += num_pairs
            self.inner_product = InnerProduct(device=device)

        if self.use_outer:
            product_out_dim += num_pairs
            self.outer_product = OuterProduct(
                inputs_dim,
                self.embedding_size,
                kernel_type=self.outer_kernel_type,
                device=device,
            )

        # dnn layer setup
        if self.use_dnn:
            # embedding layer
            self.dnn_embedding_layer = embedding_dict_gen(
                self.sparse_defs, init_std=init_std, sparse=False, device=device
            )
            self.add_regularization_weight(
                self.dnn_embedding_layer.parameters(), l2=l2_reg_embedding
            )

            self.dnn = DNN(
                product_out_dim
                + compute_inputs_dim(
                    sparse_feature_defs=self.sparse_defs,
                    dense_feature_defs=self.dense_defs,
                ),
                dnn_hidden_units,
                activation=dnn_activation,
                l2_reg=l2_reg_dnn,
                dropout_rate=dnn_dropout,
                use_bn=dnn_use_bn,
                init_std=init_std,
                device=device,
            )
            self.final_linear = nn.Linear(dnn_hidden_units[-1], 1, bias=False).to(
                device
            )

            self.add_regularization_weight(
                filter(
                    lambda x: "weight" in x[0] and "bn" not in x[0],
                    self.dnn.named_parameters(),
                ),
                l2=l2_reg_dnn,
            )
            self.add_regularization_weight(self.final_linear.weight, l2=l2_reg_dnn)

        # output layer setup
        self.output = OutputLayer(output_fn=output_fn, output_fn_args=output_fn_args)

    def forward(self, X):
        dnn_dense_vals, dnn_sparse_embs = collect_inputs_and_embeddings(
            X,
            sparse_feature_defs=self.sparse_defs,
            dense_feature_defs=self.dense_defs,
            embedding_layer_def=self.dnn_embedding_layer,
        )

        linear_signal = torch.flatten(concatenate(dnn_sparse_embs), start_dim=1)
        if self.use_inner:
            inner_product = torch.flatten(
                self.inner_product(dnn_sparse_embs), start_dim=1
            )
        if self.use_outer:
            outer_product = self.outer_product(dnn_sparse_embs)

        if self.use_outer and self.use_inner:
            product_layer = torch.cat(
                [linear_signal, inner_product, outer_product], dim=1
            )
        elif self.use_outer:
            product_layer = torch.cat([linear_signal, outer_product], dim=1)
        elif self.use_inner:
            product_layer = torch.cat([linear_signal, inner_product], dim=1)
        else:
            product_layer = linear_signal

        # dnn
        if self.use_dnn:
            dnn_input = concat_inputs([product_layer], dnn_dense_vals)
            dnn_output = self.dnn(dnn_input)
            dnn_logit = self.final_linear(dnn_output)
            logit = dnn_logit
        y = self.output(logit)
        return y



import torch.nn as nn


class Linear(nn.Module):
    def __init__(self, **kwargs):
        super(Linear, self).__init__()

    def forward(self, inputs):
        return inputs


def activation_gen(name):
    """Generate activation layers

    Args:
        name: str or nn.Module, name of activation function
    Return:
        layer: activation layer
    """
    if isinstance(name, str):
        if name.lower() == "sigmoid":
            layer = nn.Sigmoid()
        elif name.lower() == "linear":
            layer = Linear()
        elif name.lower() == "relu":
            layer = nn.ReLU(inplace=True)
        elif name.lower() == "prelu":
            layer = nn.PReLU()
    elif issubclass(name, nn.Module):
        layer = name()
    else:
        raise NotImplementedError

    return layer



import copy

import ray
from ray.train import Trainer as RayTrainer
from ray.train.callbacks import JsonLoggerCallback, TBXLoggerCallback

from veloce.train_fn import RecommendationFn
from veloce.callback import EarlyStoppingCallback
from veloce.heterogeneous import HeterogeneousStrategy, PSTrainer

callback_mapping = {"json": JsonLoggerCallback, "tbx": TBXLoggerCallback}
DefaultRun = {}


class HostedTrainer(object):
    def __init__(self, hetero_strategy, *, num_workers, use_gpu):
        self.hetero_strategy = hetero_strategy
        self.num_workers = num_workers
        self.use_gpu = use_gpu
        if self.hetero_strategy.use_raytrainer():
            self.raytrainer = RayTrainer(
                "torch", num_workers=self.num_workers, use_gpu=self.use_gpu
            )
        elif self.hetero_strategy.use_sync_ps():
            self.pstrainer = PSTrainer(
                num_workers=self.num_workers, use_gpu=self.use_gpu
            )

    def start(self):
        if self.hetero_strategy.use_raytrainer():
            self.raytrainer.start()

    def run(self, **kwargs):
        model = kwargs.pop("model")
        if self.hetero_strategy.use_raytrainer():
            result = self.raytrainer.run(train_func=RecommendationFn(model), **kwargs)
        elif self.hetero_strategy.use_sync_ps():
            result = self.pstrainer.run(
                model=model, config=kwargs["config"], dataset=kwargs["dataset"]
            )
        else:
            raise NotImplementedError
        return result

    def shutdown(self):
        if self.hetero_strategy.use_raytrainer():
            self.raytrainer.shutdown()
        else:
            ray.shutdown()


class NeuralNetTrainer(object):
    def __init__(
        self,
        module,
        module_params,
        dataset,
        dataset_options,
        batch_size,
        loss_fn,
        optimizer,
        metric_fns,
        *,
        num_workers=1,
        use_gpu=False,
        backend="torch",
        epochs=0,
        use_static_graph=False,
        optimizer_args=None,
        ddp_options=None,
        callbacks=None,
        use_early_stopping=False,
        early_stopping_args=None,
        shared_validation_dataset=None,
        heterogeneous_strategy=None,
    ):
        # current only support torch
        self.backend = backend

        self.model = module
        self.module_params = module_params
        self.ddp_options = ddp_options
        self.dataset = dataset
        self.dataset_options = dataset_options
        self.shared_validation_dataset = shared_validation_dataset

        self.epochs = epochs
        self.batch_size = batch_size
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.optimizer_args = optimizer_args or {}
        self.metric_fns = metric_fns
        self.use_static_graph = use_static_graph
        self.use_early_stopping = use_early_stopping
        self.early_stopping_args = early_stopping_args

        self.callbacks = callbacks
        self.hetero_strategy = heterogeneous_strategy or HeterogeneousStrategy()
        self.num_workers = num_workers
        self.use_gpu = use_gpu

        if self.callbacks is not None:
            self.callbacks = [callback_mapping[l]() for l in self.callbacks]
        else:
            self.callbacks = []

    def run(self, multi_runs=None, runs_mode="sync", use_checkpoint=False):
        if runs_mode != "sync":
            raise ValueError("Arg run_mode current not support non-sync")

        if multi_runs is None:
            self.multi_runs = [DefaultRun]
        else:
            self.multi_runs = multi_runs
            use_checkpoint = True
        results = []
        latest_ckpt = None

        trainer = HostedTrainer(
            self.hetero_strategy,
            num_workers=self.num_workers,
            use_gpu=self.use_gpu,
        )
        trainer.start()
        for addon in self.multi_runs:
            common_conf = dict(
                epochs=self.epochs,
                batch_size=self.batch_size,
                loss_fn=self.loss_fn,
                optimizer=self.optimizer,
                optimizer_args=self.optimizer_args,
                metric_fns=self.metric_fns,
                ddp_options=self.ddp_options,
                dataset_options=self.dataset_options,
                use_static_graph=self.use_static_graph,
                use_early_stopping=self.use_early_stopping,
                early_stopping_args=self.early_stopping_args,
                shared_validation_dataset=self.shared_validation_dataset,
                **self.module_params,
            )
            try:
                addon_module_params = addon.pop("module_params")
            except:
                addon_module_params = {}
            common_conf.update(addon)
            common_conf.update(addon_module_params)

            # early stopping part
            if not self.use_early_stopping:
                results.append(
                    trainer.run(
                        model=self.model,
                        dataset=self.dataset,
                        callbacks=self.callbacks,
                        config=common_conf,
                        checkpoint=latest_ckpt,
                    )
                )
            else:
                # hacky solution on EarlyStopping
                _es = EarlyStoppingCallback(trainer)
                _callbacks = copy.deepcopy(self.callbacks)
                _callbacks.append(_es)
                try:
                    trainer.run(
                        model=self.model,
                        dataset=self.dataset,
                        callbacks=_callbacks,
                        config=common_conf,
                        checkpoint=latest_ckpt,
                    )
                except:
                    pass
                results.append(_es.stored_results)

            # checkpoint part
            if use_checkpoint:
                latest_ckpt = trainer.latest_checkpoint
                if latest_ckpt:
                    latest_ckpt["epoch_id"] = 0

        if not self.use_early_stopping:
            trainer.shutdown()
        return results



import torch
import torch.nn as nn
import ray.train as train

from veloce.util import get_package_name, inspect_func_args
from veloce.inputs import rebuild_feature_values, find_feature_values
from veloce.model.ctr import (
    WideAndDeep as _WideAndDeep,
    DeepFM as _DeepFM,
    PNN as _PNN,
    FNN as _FNN,
)
from veloce import Epochvisor


class BaseTrainFn(object):
    def __call__(self, config):
        self.set_config(config)
        # dataset setup
        train_dataset_shard = train.get_dataset_shard("train")
        self.train_dataset_iterator = train_dataset_shard.iter_datasets()
        try:
            validation_dataset_shard = (
                self.shared_validation_dataset_shard
                or train.get_dataset_shard("validation")
            )
            self.validation_dataset_iterator = validation_dataset_shard.iter_datasets()
        except:
            self.validation_dataset_iterator = None

    def set_config(self, config):
        # basic configs setup
        self.dataset_options = config.get("dataset_options", None)
        self.seed = config.get("seed", 1024)
        self.epochs = config.get("epochs", 10)
        self.batch_size = config.get("batch_size", 256)
        self.loss_fn = config.get("loss_fn", nn.BCELoss())
        self.optimizer = config.get("optimizer", torch.optim.Adam)
        self.optimizer_args = config.get("optimizer_args", None) or {}
        self.output_fn = config.get("output_fn", torch.sigmoid)
        self.output_fn_args = config.get("output_fn_args", None)
        self.summary_nn_arch = config.get("summary_nn_arch", False)
        self.init_std = config.get("init_std", 0.0001)
        self.ddp_options = config.get("ddp_options", None) or {}
        self.use_static_graph = config.get("use_static_graph", False)
        self.use_early_stopping = config.get("use_early_stopping", False)
        self.early_stopping_args = config.get("early_stopping_args", None) or {}
        self.shared_validation_dataset_shard = config.get(
            "shared_validation_dataset", None
        )
        try:
            import torchmetrics

            self.metric_fns = config.get("metric_fns", [torchmetrics.AUROC()])
        except:
            self.metric_fns = config.get("metric_fns", None)
            if not self.metric_fns:
                raise ValueError(
                    "Arg metric_fns must be given and valid, you may use sklearn or torchmetrics metric functions"
                )
        try:
            self.checkpoint = train.load_checkpoint() or None
            self.device = train.torch.get_device()
        except:
            self.checkpoint = None
            self.device = config.get("device", "cpu")

    def setup_model(
        self, model_args, model=None, enable_optimizer=True, enable_ddp=True
    ):
        # model setup
        if (not hasattr(self, "model") or self.model is None) and model is not None:
            self.model = model
        if enable_ddp:
            self.model = train.torch.prepare_model(
                model=self.model(**model_args), ddp_kwargs=self.ddp_options
            )
        else:
            self.model = self.model(**model_args)
        if self.use_static_graph:
            self.model._set_static_graph()
        if self.summary_nn_arch:
            print(self.model)

        if enable_optimizer:
            # optimizer setup
            if get_package_name(self.optimizer) == "torch" or (
                get_package_name(self.optimizer) == "veloce"
                and type(self.optimizer).__name__ != "OptimizerStack"
            ):
                self.optimizer = self.optimizer(
                    self.model.parameters(), **self.optimizer_args
                )
            elif type(self.optimizer).__name__ == "OptimizerStack":
                self.optimizer.compile(self.model.module if enable_ddp else self.model)
            else:
                raise ValueError("Arg optimizer must be given and valid")

    def setup_epv(self):
        self.epv = Epochvisor(
            epochs=self.epochs,
            train_dataset_iter=self.train_dataset_iterator,
            validation_dataset_iter=self.validation_dataset_iterator,
            dataset_options=self.dataset_options,
            batch_size=self.batch_size,
            model=self.model,
            loss_fn=self.loss_fn,
            optimizer=self.optimizer,
            metric_fns=self.metric_fns,
            device=self.device,
            checkpoint=self.checkpoint,
            use_early_stopping=self.use_early_stopping,
            early_stopping_args=self.early_stopping_args,
        )

    def run_epochs(self):
        return self.epv.run_epochs()


class RecommendationFn(BaseTrainFn):
    def __init__(self, model):
        self.model = model

    def __call__(self, config):
        super(RecommendationFn, self).__call__(config)
        for k in config.keys():
            if find_feature_values(config[k]):
                config[k] = rebuild_feature_values(config[k])
        model_args = inspect_func_args(config, self.model)
        self.setup_model(model_args=model_args)
        self.setup_epv()
        return self.run_epochs()


# native train fns
WideAndDeep = RecommendationFn(_WideAndDeep)
DeepFM = RecommendationFn(_DeepFM)
PNN = RecommendationFn(_PNN)
FNN = RecommendationFn(_FNN)



class LossFnStack(object):
    def __init__(self, *loss_fns):
        if sum([fn_wrap["weight"] for fn_wrap in loss_fns]) != 1:
            raise ValueError("The sum of all weights in LossFnStack must be 1")
        self.loss_fns = loss_fns

    def __call__(self, pred, y):
        return sum(
            [fn_wrap["weight"] * fn_wrap["fn"](pred, y) for fn_wrap in self.loss_fns]
        )



import torch
import torch.nn as nn


class FM(nn.Module):
    def __init__(self):
        super(FM, self).__init__()

    def forward(self, inputs):
        square_of_sum = torch.pow(torch.sum(inputs, dim=1), 2)
        sum_of_square = torch.sum(inputs * inputs, dim=1)
        cross_term = square_of_sum - sum_of_square
        cross_term = 0.5 * torch.sum(cross_term, dim=1, keepdim=True)
        return cross_term


class FMNative(nn.Module):
    def __init__(self, feature_def_dims, k_factor, dropout_rate, init_std=0.0001):
        super(FMNative, self).__init__()
        self.n = feature_def_dims
        self.k = k_factor
        self.linear = nn.Linear(self.n, 1, bias=True)
        self.v = nn.Parameter(torch.rand(self.n, self.k))
        torch.nn.init.normal_(self.v, mean=0, std=init_std)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, inputs):
        fm_first = self.linear(inputs)
        fm_second_part1 = torch.pow(torch.matmul(inputs, self.v), 2)
        fm_second_part2 = torch.matmul(torch.pow(inputs, 2), torch.pow(self.v, 2))
        fm_second = 0.5 * torch.sum(
            fm_second_part1 - fm_second_part2, dim=1, keepdim=True
        )
        fm_second = self.dropout(fm_second)
        return fm_first + fm_second


class InnerProduct(nn.Module):
    def __init__(self, reduce_sum=True, device="cpu"):
        super(InnerProduct, self).__init__()
        self.reduce_sum = reduce_sum
        self.to(device)

    def forward(self, inputs):
        embed_list = inputs
        row = []
        col = []
        num_inputs = len(embed_list)
        for i in range(num_inputs - 1):
            for j in range(i + 1, num_inputs):
                row.append(i)
                col.append(j)
        p = torch.cat([embed_list[idx] for idx in row], dim=1)  # batch num_pairs k
        q = torch.cat([embed_list[idx] for idx in col], dim=1)
        inner_product = p * q
        if self.reduce_sum:
            inner_product = torch.sum(inner_product, dim=2, keepdim=True)
        return inner_product


class OuterProduct(nn.Module):
    def __init__(
        self, field_size, embedding_size, kernel_type="mat", seed=1024, device="cpu"
    ):
        super(OuterProduct, self).__init__()
        self.kernel_type = kernel_type

        num_inputs = field_size
        num_pairs = int(num_inputs * (num_inputs - 1) / 2)
        embed_size = embedding_size
        if self.kernel_type == "mat":

            self.kernel = nn.Parameter(torch.Tensor(embed_size, num_pairs, embed_size))

        elif self.kernel_type == "vec":
            self.kernel = nn.Parameter(torch.Tensor(num_pairs, embed_size))

        elif self.kernel_type == "num":
            self.kernel = nn.Parameter(torch.Tensor(num_pairs, 1))
        nn.init.xavier_uniform_(self.kernel)
        self.to(device)

    def forward(self, inputs):
        embed_list = inputs
        row = []
        col = []
        num_inputs = len(embed_list)
        for i in range(num_inputs - 1):
            for j in range(i + 1, num_inputs):
                row.append(i)
                col.append(j)
        p = torch.cat([embed_list[idx] for idx in row], dim=1)  # batch num_pairs k
        q = torch.cat([embed_list[idx] for idx in col], dim=1)

        if self.kernel_type == "mat":
            p.unsqueeze_(dim=1)
            kp = torch.sum(
                torch.mul(
                    torch.transpose(torch.sum(torch.mul(p, self.kernel), dim=-1), 2, 1),
                    q,
                ),
                dim=-1,
            )
        else:
            k = torch.unsqueeze(self.kernel, 0)
            kp = torch.sum(p * q * k, dim=-1)
        return kp



import torch
import torch.nn as nn

from veloce.inputs import embedding_dict_gen, collect_inputs_and_embeddings


class BaseModel(nn.Module):
    def __init__(self, **kwargs):
        super(BaseModel, self).__init__()
        torch.manual_seed(kwargs["seed"])
        self.device = kwargs["device"]
        self.regularization_weight = []
        self.to(self.device)

    def add_regularization_weight(self, weight_list, l1=0.0, l2=0.0):
        if isinstance(weight_list, torch.nn.parameter.Parameter):
            weight_list = [weight_list]
        else:
            weight_list = list(weight_list)
        self.regularization_weight.append((weight_list, l1, l2))

    def get_regularization_loss(self):
        total_reg_loss = torch.zeros((1,), device=self.device)
        for weight_list, l1, l2 in self.regularization_weight:
            for w in weight_list:
                if isinstance(w, tuple):
                    parameter = w[1]
                else:
                    parameter = w
                if l1 > 0:
                    total_reg_loss += torch.sum(l1 * torch.abs(parameter))
                if l2 > 0:
                    try:
                        total_reg_loss += torch.sum(l2 * torch.square(parameter))
                    except AttributeError:
                        total_reg_loss += torch.sum(l2 * parameter * parameter)
        return total_reg_loss


class Linear(nn.Module):
    def __init__(
        self,
        sparse_feature_defs,
        dense_feature_defs,
        init_std=0.0001,
        device="cpu",
    ):
        super(Linear, self).__init__()
        self.device = device
        self.sparse_feature_defs = sparse_feature_defs
        self.dense_feature_defs = dense_feature_defs

        self.linear_embedding_dict = embedding_dict_gen(
            self.sparse_feature_defs, init_std, linear=True, sparse=False, device=device
        )

        if len(self.dense_feature_defs):
            self.weight = nn.Parameter(
                torch.Tensor(sum(fc.dimension for fc in self.dense_feature_defs), 1).to(
                    device
                )
            )
            torch.nn.init.normal_(self.weight, mean=0, std=init_std)

    def forward(self, X, sparse_feat_refine_weight=None):
        dense_values, sparse_embeddings = collect_inputs_and_embeddings(
            X,
            sparse_feature_defs=self.sparse_feature_defs,
            dense_feature_defs=self.dense_feature_defs,
            embedding_layer_def=self.linear_embedding_dict,
        )

        linear_logit = torch.zeros([X.shape[0], 1]).to(self.device)
        if len(sparse_embeddings) > 0:
            linear_logit = linear_logit.to(sparse_embeddings[0].device)
            sparse_embedding_cat = torch.cat(sparse_embeddings, dim=-1)
            if sparse_feat_refine_weight is not None:
                sparse_embedding_cat = (
                    sparse_embedding_cat * sparse_feat_refine_weight.unsqueeze(1)
                )
            sparse_feat_logit = torch.sum(sparse_embedding_cat, dim=-1, keepdim=False)
            linear_logit += sparse_feat_logit
        if len(dense_values) > 0:
            dense_value_logit = torch.cat(dense_values, dim=-1).matmul(self.weight)
            linear_logit += dense_value_logit
        return linear_logit



import torch.nn as nn
import torch

from veloce.layer import DNN, OutputLayer, FMNative
from veloce.inputs import (
    concat_inputs,
    compute_inputs_dim,
    embedding_dict_gen,
    collect_inputs_and_embeddings,
)
from .base import BaseModel, Linear


class WideAndDeep(BaseModel):
    def __init__(
        self,
        # feature defs
        dense_feature_defs=None,
        sparse_feature_defs=None,
        # wide related
        use_fm=False,
        k_factor=10,
        fm_dropout=0,
        l2_reg_linear=1e-3,
        # dnn related
        dnn_hidden_units=(256, 128),
        dnn_use_bn=False,
        dnn_activation="relu",
        dnn_dropout=0,
        l2_reg_embedding=1e-3,
        l2_reg_dnn=1e-3,
        # base configs
        seed=1024,
        output_fn=torch.sigmoid,
        output_fn_args=None,
        device="cpu",
        init_std=1e-4,
    ):
        super(WideAndDeep, self).__init__(seed=seed, device=device)
        self.dense_defs = dense_feature_defs
        self.sparse_defs = sparse_feature_defs
        self.use_dnn = len(dnn_hidden_units) > 0
        self.use_fm = use_fm

        # embedding layer setup
        self.dnn_embedding_layer = embedding_dict_gen(
            self.sparse_defs, init_std=init_std, sparse=False, device=device
        )
        self.add_regularization_weight(
            self.dnn_embedding_layer.parameters(), l2=l2_reg_embedding
        )

        # wide model setup
        if self.use_fm:
            self.wide_model = FMNative(
                feature_def_dims=sum(fc.dimension for fc in self.dense_defs)
                + sum(fc.embedding_dim for fc in self.sparse_defs),
                k_factor=k_factor,
                dropout_rate=fm_dropout,
                init_std=init_std,
            )
        else:
            self.wide_model = Linear(
                sparse_feature_defs=self.sparse_defs,
                dense_feature_defs=self.dense_defs,
                device=device,
            )
        self.add_regularization_weight(self.wide_model.parameters(), l2=l2_reg_linear)

        # deep model setup
        if self.use_dnn:
            self.deep_model = DNN(
                compute_inputs_dim(
                    sparse_feature_defs=self.sparse_defs,
                    dense_feature_defs=self.dense_defs,
                ),
                dnn_hidden_units,
                activation=dnn_activation,
                l2_reg=l2_reg_dnn,
                dropout_rate=dnn_dropout,
                use_bn=dnn_use_bn,
                init_std=init_std,
                device=device,
            )
            self.final_linear = nn.Linear(dnn_hidden_units[-1], 1, bias=False).to(
                device
            )

            self.add_regularization_weight(
                filter(
                    lambda x: "weight" in x[0] and "bn" not in x[0],
                    self.deep_model.named_parameters(),
                ),
                l2=l2_reg_dnn,
            )
            self.add_regularization_weight(self.final_linear.weight, l2=l2_reg_dnn)

        # output layer setup
        self.output = OutputLayer(output_fn=output_fn, output_fn_args=output_fn_args)

    def forward(self, X):
        if self.use_fm:
            dense_values, sparse_embeddings = collect_inputs_and_embeddings(
                X,
                sparse_feature_defs=self.sparse_defs,
                dense_feature_defs=self.dense_defs,
                embedding_layer_def=self.dnn_embedding_layer,
            )
            wide_input = concat_inputs(sparse_embeddings, dense_values)
            logit = self.wide_model(wide_input)
        else:
            logit = self.wide_model(X)
        if self.use_dnn:
            dense_values, sparse_embeddings = collect_inputs_and_embeddings(
                X,
                sparse_feature_defs=self.sparse_defs,
                dense_feature_defs=self.dense_defs,
                embedding_layer_def=self.dnn_embedding_layer,
            )
            dnn_input = concat_inputs(sparse_embeddings, dense_values)
            dnn_output = self.deep_model(dnn_input)
            dnn_logit = self.final_linear(dnn_output)
            logit += dnn_logit
        y = self.output(logit)
        return y



import torch.nn as nn
import torch

from veloce.layer import DNN, OutputLayer, FM
from veloce.inputs import (
    concat_inputs,
    compute_inputs_dim,
    embedding_dict_gen,
    collect_inputs_and_embeddings,
)
from .base import BaseModel, Linear


class DeepFM(BaseModel):
    def __init__(
        self,
        # feature_defs
        dense_feature_defs=None,
        sparse_feature_defs=None,
        # fm related
        use_fm=True,
        l2_reg_fm_1=1e-3,
        l2_reg_fm_2=1e-3,
        # dnn related
        dnn_hidden_units=(256, 128),
        dnn_use_bn=False,
        dnn_activation="relu",
        l2_reg_embedding=1e-3,
        l2_reg_dnn=1e-3,
        dnn_dropout=0,
        # base config
        seed=1024,
        output_fn=torch.sigmoid,
        output_fn_args=None,
        device="cpu",
        init_std=1e-4,
    ):
        super(DeepFM, self).__init__(seed=seed, device=device)
        self.dense_defs = dense_feature_defs
        self.sparse_defs = sparse_feature_defs
        self.use_dnn = len(dnn_hidden_units) > 0
        self.use_fm = use_fm

        # fm layers setup
        if self.use_fm:
            self.fm_1 = Linear(
                sparse_feature_defs=self.sparse_defs,
                dense_feature_defs=self.dense_defs,
                device=device,
            )
            self.add_regularization_weight(self.fm_1.parameters(), l2=l2_reg_fm_1)

            self.fm_2 = FM()
            # fm_2 embedding layer
            self.fm_2_embedding_layer = embedding_dict_gen(
                self.sparse_defs,
                init_std=init_std,
                sparse=False,
                device=device,
            )
            self.add_regularization_weight(
                self.fm_2_embedding_layer.parameters(), l2=l2_reg_fm_2
            )

        # dnn layer setup
        if self.use_dnn:
            # embedding layer
            self.dnn_embedding_layer = embedding_dict_gen(
                self.sparse_defs, init_std=init_std, sparse=False, device=device
            )
            self.add_regularization_weight(
                self.dnn_embedding_layer.parameters(), l2=l2_reg_embedding
            )

            self.dnn = DNN(
                compute_inputs_dim(
                    sparse_feature_defs=self.sparse_defs,
                    dense_feature_defs=self.dense_defs,
                ),
                dnn_hidden_units,
                activation=dnn_activation,
                l2_reg=l2_reg_dnn,
                dropout_rate=dnn_dropout,
                use_bn=dnn_use_bn,
                init_std=init_std,
                device=device,
            )
            self.final_linear = nn.Linear(dnn_hidden_units[-1], 1, bias=False).to(
                device
            )

            self.add_regularization_weight(
                filter(
                    lambda x: "weight" in x[0] and "bn" not in x[0],
                    self.dnn.named_parameters(),
                ),
                l2=l2_reg_dnn,
            )
            self.add_regularization_weight(self.final_linear.weight, l2=l2_reg_dnn)

        # output layer setup
        self.output = OutputLayer(output_fn=output_fn, output_fn_args=output_fn_args)

    def forward(self, X):
        if self.use_fm:
            # fm_1
            logit = self.fm_1(X)
            # fm_2
            _, fm_2_sparse_embs = collect_inputs_and_embeddings(
                X,
                sparse_feature_defs=self.sparse_defs,
                embedding_layer_def=self.fm_2_embedding_layer,
            )
            fm_2_input = torch.cat(fm_2_sparse_embs, dim=1)
            logit += self.fm_2(fm_2_input)

        # dnn
        if self.use_dnn:
            dnn_dense_vals, dnn_sparse_embs = collect_inputs_and_embeddings(
                X,
                sparse_feature_defs=self.sparse_defs,
                dense_feature_defs=self.dense_defs,
                embedding_layer_def=self.dnn_embedding_layer,
            )
            dnn_input = concat_inputs(dnn_sparse_embs, dnn_dense_vals)
            dnn_output = self.dnn(dnn_input)
            dnn_logit = self.final_linear(dnn_output)
            logit += dnn_logit
        y = self.output(logit)
        return y



#!/usr/bin/env python

import os
import re
from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(__file__)

def find_version(*filepath):
    # Extract version information from filepath
    with open(os.path.join(ROOT_DIR, *filepath)) as fp:
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", fp.read(), re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="veloce",
    version=find_version("veloce", "version.py"),
    author="Xiaoyu Zhai",
    author_email="xiaoyu.zhai@hotmail.com",
    description="Veloce: An instant distributed computing library based on Ray stack",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ryantd/veloce",
    packages=find_packages(),
    package_dir={
        "veloce": "veloce",
        "docs": "docs",
    },
    install_requires=[
        "torch>=1.9.1",
        "ray>=1.9.2,<=1.10",
        "pyarrow>=6.0.1",
        "pandas>=1.3.5,<=1.4.1",
        "requests>=2.23.0",
    ],
    python_requires=">=3.7.1",
)



import time
from collections import defaultdict

import torch
import ray.train as train

from .util import get_package_name, merge_results, get_type, get_func_name


class LossMetricAccumulator(object):
    def __init__(self, loss_fn, metric_fns):
        self.sklearn_intermediates = defaultdict(int)
        self.torchm_intermediates = defaultdict(int)
        self.loss = 0
        self.loss_fn = loss_fn
        self.metric_fns = metric_fns

    def acc(self, pred, y):
        self.loss += self.loss_fn(pred, y).item()
        for fn in self.metric_fns:
            if get_package_name(fn) == "torchmetrics":
                r = fn(pred, y.int())
                if get_type(fn) != "function":
                    continue
                self.torchm_intermediates[get_func_name(fn)] += r
            elif get_package_name(fn) == "sklearn":
                self.sklearn_intermediates[get_func_name(fn)] += fn(
                    y.cpu().data.numpy(), pred.cpu().data.numpy()
                )

    def gen_result(self, num_batches, with_latest_loss=False):
        results = {get_type(self.loss_fn): self.loss / num_batches}
        # torchmetrics module compute and reset
        for fn in self.metric_fns:
            if get_package_name(fn) == "torchmetrics" and get_type(fn) != "function":
                results[get_type(fn)] = fn.compute().item()
                fn.reset()
        # torchmetrics functional compute
        for name in self.torchm_intermediates.keys():
            self.torchm_intermediates[name] /= num_batches
        # sklearn compute
        for name in self.sklearn_intermediates.keys():
            self.sklearn_intermediates[name] /= num_batches
        results.update(self.torchm_intermediates)
        results.update(self.sklearn_intermediates)

        if with_latest_loss:
            return results, self.loss / num_batches
        return results


class Epochvisor(object):
    def __init__(
        self,
        epochs,
        *,
        dataset_options,
        batch_size,
        model,
        loss_fn,
        optimizer,
        device,
        train_dataset_iter,
        checkpoint=None,
        metric_fns=None,
        validation_dataset_iter=None,
        use_early_stopping=False,
        early_stopping_args=None,
    ):
        self.epochs = epochs
        self.train_dataset_iter = train_dataset_iter
        self.validation_dataset_iter = validation_dataset_iter
        self.dataset_options = dataset_options
        self.batch_size = batch_size
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.metric_fns = metric_fns
        self.device = device
        self.checkpoint = checkpoint
        self.use_early_stopping = use_early_stopping
        self.early_stopping_args = early_stopping_args or {"patience": 2}

        if not self.metric_fns:
            raise ValueError("metric_fns must be given")

    def run_epochs(self):
        label_column = self.dataset_options["label_column"]
        feature_columns = self.dataset_options["feature_columns"]
        label_column_dtype = self.dataset_options["label_column_dtype"]
        feature_column_dtypes = self.dataset_options["feature_column_dtypes"]
        start_epoch = 0
        results = []
        is_checkpoint_based = False

        # early stopping related
        previous_epoch_loss = 1
        es_trigger_times = 0
        is_early_stopped = False

        if self.checkpoint:
            is_checkpoint_based = True
            model_state_dict = self.checkpoint.get("model_state_dict", None)
            optimizer_state_dict = self.checkpoint.get("optimizer_state_dict", None)
            start_epoch = self.checkpoint.get("epoch_id", 0)

            self.model.load_state_dict(model_state_dict)
            self.optimizer.load_state_dict(optimizer_state_dict)

        for epoch_id in range(start_epoch, self.epochs):
            start_ts = time.time()
            train_dataset = next(self.train_dataset_iter)
            train_torch_dataset = train_dataset.to_torch(
                label_column=label_column,
                feature_columns=feature_columns,
                label_column_dtype=label_column_dtype,
                feature_column_dtypes=feature_column_dtypes,
                batch_size=self.batch_size,
            )
            train_result = self.train_epoch(train_torch_dataset)
            if self.validation_dataset_iter:
                validation_dataset = next(self.validation_dataset_iter)
                validation_torch_dataset = validation_dataset.to_torch(
                    label_column=label_column,
                    feature_columns=feature_columns,
                    label_column_dtype=label_column_dtype,
                    feature_column_dtypes=feature_column_dtypes,
                    batch_size=self.batch_size,
                )
                validation_result, latest_epoch_loss = self.validate_epoch(
                    validation_torch_dataset, with_latest_loss=True
                )
            end_ts = time.time()

            # early stopping check
            if self.use_early_stopping:
                if latest_epoch_loss > previous_epoch_loss:
                    es_trigger_times += 1
                    if (
                        es_trigger_times >= self.early_stopping_args["patience"]
                        and epoch_id != self.epochs - 1
                    ):
                        is_early_stopped = True
                else:
                    es_trigger_times = 0

            # post epoch operations
            result = merge_results(
                validation_result=validation_result,
                train_result=train_result,
                time_diff=end_ts - start_ts,
                is_early_stopped=is_early_stopped,
                is_checkpoint_based=is_checkpoint_based,
            )
            train.save_checkpoint(
                epoch_id=epoch_id,
                model_state_dict=self.model.state_dict(),
                optimizer_state_dict=self.optimizer.state_dict(),
            )
            train.report(**result)
            results.append(result)
            previous_epoch_loss = latest_epoch_loss
        return results

    def train_epoch(self, train_iterable_ds):
        self.model.train()
        num_batches = 0
        lma = LossMetricAccumulator(loss_fn=self.loss_fn, metric_fns=self.metric_fns)
        for X, y in train_iterable_ds:
            X = X.to(self.device)
            y = y.to(self.device)
            pred = self.model(X)
            if hasattr(self.model.module, "get_regularization_loss"):
                loss = (
                    self.loss_fn(pred, y) + self.model.module.get_regularization_loss()
                )
            else:
                loss = self.loss_fn(pred, y)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            # metrics part
            num_batches += 1
            lma.acc(pred=pred, y=y)
        return lma.gen_result(num_batches=num_batches)

    def validate_epoch(self, validation_iterable_ds, with_latest_loss=False):
        self.model.eval()
        num_batches = 0
        lma = LossMetricAccumulator(loss_fn=self.loss_fn, metric_fns=self.metric_fns)
        with torch.no_grad():
            for X, y in validation_iterable_ds:
                X = X.to(self.device)
                y = y.to(self.device)
                pred = self.model(X)
                # metrics part
                num_batches += 1
                lma.acc(pred=pred, y=y)
        return lma.gen_result(
            num_batches=num_batches, with_latest_loss=with_latest_loss
        )



from typing import List

import pandas as pd

from veloce.inputs import SparseFeat, DenseFeat


class SparseFeatureDef(object):
    def __init__(
        self,
        col_selectors: List,
        vocabulary_size_fn=lambda x, y: x[y].max() + 1,
        embedding_dim=4,
    ):
        self.col_selectors = col_selectors
        self.embedding_dim = embedding_dim
        self.vocabulary_size_fn = vocabulary_size_fn

    def __call__(self, data):
        if isinstance(data, pd.DataFrame):
            return self._call_pandas_dataframe(data)
        else:
            raise NotImplementedError

    def _call_pandas_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        sparse_df = pd.DataFrame()
        for col in self.col_selectors:
            sparse_feat = SparseFeat(
                col,
                vocabulary_size=self.vocabulary_size_fn(df, col),
                column_idx=df.columns.get_loc(col) - 1,
                embedding_dim=self.embedding_dim,
            )
            sparse_df = sparse_df.append(sparse_feat.__dict__, ignore_index=True)
        return sparse_df


class DenseFeatureDef(object):
    def __init__(self, col_selectors: List, dimension=1):
        self.col_selectors = col_selectors
        self.dimension = dimension

    def __call__(self, df):
        if isinstance(df, pd.DataFrame):
            return self._call_pandas_dataframe(df)
        else:
            raise NotImplementedError

    def _call_pandas_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        dense_df = pd.DataFrame()
        for col in self.col_selectors:
            dense_feat = DenseFeat(
                col,
                column_idx=df.columns.get_loc(col) - 1,
                dimension=self.dimension,
            )
            dense_df = dense_df.append(dense_feat.__dict__, ignore_index=True)
        return dense_df
